<?php

namespace App\Controller;

use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Security\Http\Attribute\IsGranted;
use Symfony\Component\Serializer\Encoder\JsonEncoder;
use Symfony\Component\Serializer\Normalizer\ObjectNormalizer;
use Symfony\Component\Serializer\Serializer;
use Doctrine\Common\Annotations\AnnotationReader;
use Symfony\Component\Serializer\Mapping\Factory\ClassMetadataFactory;
use Symfony\Component\Serializer\Mapping\Loader\AnnotationLoader;
use App\Entity\Transaction;
use App\Entity\User;
use App\Service\CategoryService;
use Doctrine\Persistence\ManagerRegistry;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\PropertyInfo\Extractor\ReflectionExtractor;
use Symfony\Component\Serializer\Normalizer\DateTimeNormalizer;

#[IsGranted('ROLE_USER')]
class ApiController extends AbstractController
{
    private Serializer $serializer;
    private CategoryService $categoryService;

    public function __construct(CategoryService $categoryService)
    {
        $classMetadataFactory = new ClassMetadataFactory(new AnnotationLoader(new AnnotationReader()));

        $dateNormalizer = new DateTimeNormalizer([DateTimeNormalizer::FORMAT_KEY => "Y-m-d"]);

        $encoders = [new JsonEncoder()];
        $normalizers = [$dateNormalizer, new ObjectNormalizer($classMetadataFactory, null, null, new ReflectionExtractor())];

        $this->serializer = new Serializer($normalizers, $encoders);

        $this->categoryService = $categoryService;
    }

    private function manageCategory(Transaction $transaction)
    {
        if ($transaction->getCategory() !== "auto-detect") return;
        $category = $this->categoryService->getCategory($transaction->getLocation());
        $transaction->setCategory($category);
    }

    private function checkTransactionAccess(Transaction $transaction)
    {
        /**
         * @var User
         */
        $user = $this->getUser();
        if (!$user->getTransactions()->contains($transaction)) throw $this->createNotFoundException();
    }

    
    #[Route('/api/transaction', name: 'api_transaction_list', methods: ['GET'])]
    public function list(): Response
    {
        /**
         * @var User
         */
        $user = $this->getUser();
        $transactions = $user->getTransactions();
        
        return new Response($this->serializer->serialize($transactions, "json"), 200, ["Content-Type" => "application/json"]);
    }

    #[Route("/api/transaction/{id}", name: 'api_transaction_get', methods: ['GET'])]
    public function get(Transaction $transaction): Response
    {
        $this->checkTransactionAccess($transaction);

        return new Response($this->serializer->serialize($transaction, "json"), 200, ["Content-Type" => "application/json"]);
    }

    #[Route('/api/transaction', name: 'api_transaction_create', methods: ['POST'])]
    public function create(Request $request, ManagerRegistry $doctrine): Response
    {
        $entityManager = $doctrine->getManager();
        /**
         * @var User
         */
        $user = $this->getUser();
        try {
            /**
             * @var Transaction
             */
            $transaction = $this->serializer->deserialize($request->getContent(), Transaction::class, "json", [DateTimeNormalizer::FORMAT_KEY => "Y-m-d"]);
        } catch (\Exception $error) {
            return new Response($error, 400);
        }
        $this->manageCategory($transaction);
        $user->addTransaction($transaction);
        $entityManager->persist($transaction);
        $entityManager->flush();
        return new Response($this->serializer->serialize($transaction, "json"), 200, ["Content-Type" => "application/json"]);
    }

    #[Route('/api/transaction/{id}', name: 'api_transaction_delete', methods: ['DELETE'])]
    public function delete(Transaction $transaction, ManagerRegistry $doctrine)
    {
        $this->checkTransactionAccess($transaction);
        $entityManager = $doctrine->getManager();

        $entityManager->remove($transaction);
        $entityManager->flush();

        return new Response();
    }

    #[Route('/api/transaction/{id}', name: 'api_transaction_update', methods: ['PATCH'])]
    public function update(Request $request, Transaction $transaction, ManagerRegistry $doctrine)
    {
        $this->checkTransactionAccess($transaction);
        $entityManager = $doctrine->getManager();

        try {
            $transaction = $this->serializer->deserialize($request->getContent(), Transaction::class, "json", [DateTimeNormalizer::FORMAT_KEY => "Y-m-d", "object_to_populate" => $transaction]);
        } catch (\Exception $error) {
            return new Response($error, 400);
        }

        $this->manageCategory($transaction);

        $entityManager->flush();

        return new JsonResponse($transaction->getCategory());
    }
}
